/*====================================================================
| Version: July 6, 2011
\===================================================================*/

/*====================================================================
| Philippe Thevenaz
| EPFL/STI/IMT/LIB/BM.4.137
| Station 17
| CH-1015 Lausanne VD
| Switzerland
|
| phone (CET): +41(21)693.51.61
| fax: +41(21)693.37.01
| RFC-822: philippe.thevenaz@epfl.ch
| X-400: /C=ch/A=400net/P=switch/O=epfl/S=thevenaz/G=philippe/
| URL: http://bigwww.epfl.ch/
\===================================================================*/

// ovusculeSnake2D
import ovusculeSnake2D.OvusculeSnake2D;
import ovusculeSnake2D.OvusculeSnake2DKeeper;
import ovusculeSnake2D.OvusculeSnake2DNode;
import ovusculeSnake2D.OvusculeSnake2DScale;

// ImageJ
import ij.IJ;
import ij.ImagePlus;
import ij.Macro;
import ij.plugin.filter.PlugInFilter;
import ij.process.ImageProcessor;

// Java
import java.awt.geom.Point2D;
import java.io.IOException;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.Vector;

// Java 1.5
import static java.io.StreamTokenizer.TT_EOF;
import static java.io.StreamTokenizer.TT_EOL;
import static java.io.StreamTokenizer.TT_NUMBER;
import static java.io.StreamTokenizer.TT_WORD;
import static java.lang.Math.PI;
import static java.lang.Math.abs;
import static java.lang.Math.ceil;
import static java.lang.Math.cos;
import static java.lang.Math.floor;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.lang.Math.round;
import static java.lang.Math.sin;
import static java.lang.Math.sqrt;

/*====================================================================
|	Ovuscule_
\===================================================================*/
public class Ovuscule_
	implements
		PlugInFilter

{ /* begin class Ovuscule_ */

/*....................................................................
	private variables
....................................................................*/
private ImagePlus imp = null;
private String commandLine = null;
private static final double GOLDEN_RATIO = 0.5 + sqrt(1.25);
private static final double SQRT2 = sqrt(2.0);

/*....................................................................
	PlugInFilter methods
....................................................................*/
/*------------------------------------------------------------------*/
public void run (
	final ImageProcessor ip
) {
	Runtime.getRuntime().gc();
	final OvusculeSnake2DKeeper keeper = new OvusculeSnake2DKeeper();
	String options = Macro.getOptions();
	if (!commandLine.equals("")) {
		options = commandLine;
	}
	if (null == options) {
		final int height = ip.getHeight();
		final int width = ip.getWidth();
		final double dx = max(3.0 + 2.0 * SQRT2,
			0.25 * sqrt(0.75) * (GOLDEN_RATIO - 1.0)
			* min((double)width, (double)height));
		final double dy = max(3.0 + 2.0 * SQRT2,
			0.25 * (GOLDEN_RATIO - 1.0)
			* min((double)width, (double)height));
		final double p1 = 0.5 * (double)(width - 1);
		final double p2 = 0.5 * (double)(height - 1) + dy;
		final double q1 = 0.5 * (double)(width - 1) + dx;
		final double q2 = 0.5 * (double)(height - 1) - 0.5 * dy;
		final double r1 = 0.5 * (double)(width - 1) - dx;
		final double r2 = 0.5 * (double)(height - 1) - 0.5 * dy;
		final Ovuscule ovuscule = new Ovuscule(ip, p1, p2, q1, q2, r1, r2);
		keeper.interactAndOptimize(ovuscule, imp);
	}
	else {
		final String[] token = getTokens(options);
		if ((token.length < 12) || (0 != (token.length % 2))) {
			dumpSyntax(options);
			IJ.error("Invalid syntax");
			return;
		}
		if (!(token[0].equals("x0") && token[2].equals("y0")
			&& token[4].equals("x1") && token[6].equals("y1")
			&& token[8].equals("x2") && token[10].equals("y2"))) {
			dumpSyntax(options);
			IJ.error("Invalid syntax");
			return;
		}
		boolean optimizing = true;
		boolean stamping = false;
		for (int t = 13; (t < token.length); t += 2) {
			if (token[t - 1].equals("optimizing")) {
				if (token[t].equals("no") || token[t].equals("yes")) {
					optimizing = token[t].equals("yes");
					continue;
				}
				else {
					dumpSyntax(options);
					IJ.error("Invalid syntax");
					return;
				}
			}
			if (token[t - 1].equals("stamping")) {
				if (token[t].equals("no") || token[t].equals("yes")) {
					stamping = token[t].equals("yes");
					continue;
				}
				else {
					dumpSyntax(options);
					IJ.error("Invalid syntax");
					return;
				}
			}
			dumpSyntax(options);
			IJ.error("Invalid syntax");
			return;
		}
		try {
			double p1 = new Double(token[1]).doubleValue();
			double p2 = new Double(token[3]).doubleValue();
			double q1 = new Double(token[5]).doubleValue();
			double q2 = new Double(token[7]).doubleValue();
			double r1 = new Double(token[9]).doubleValue();
			double r2 = new Double(token[11]).doubleValue();
			final Ovuscule ovuscule = new Ovuscule(ip, p1, p2, q1, q2, r1, r2);
			keeper.optimize(ovuscule, imp, optimizing, stamping);
		} catch (NumberFormatException e) {
			dumpSyntax(options);
			IJ.log("Number format exception " + e.getMessage());
			IJ.error("Invalid syntax");
			return;
		}
	}
} /* end run */

/*------------------------------------------------------------------*/
public int setup (
	final String commandLine,
	final ImagePlus imp
) {
	this.commandLine = commandLine;
	this.imp = imp;
	return(DOES_ALL);
} /* end setup */

/*....................................................................
	private methods
....................................................................*/
/*------------------------------------------------------------------*/
private void dumpSyntax (
	final String options
) {
	IJ.log("");
	IJ.log("!!!");
	IJ.log("");
	IJ.log("Invalid options: \"" + options + "\"");
	IJ.log("");
	IJ.log("---");
	IJ.log("");
	IJ.log("run(\"Ovuscule \", \"x0=<NUMBER> y0=<NUMBER> "
		+ "x1=<NUMBER> y1=<NUMBER> x2=<NUMBER> y2=<NUMBER> "
		+ "{optimizing=[yes|no]} {stamping=[yes|no]}\");");
	IJ.log("");
	IJ.log("---");
	IJ.log("");
	IJ.log("Avoid exponential notations for <NUMBER>");
	IJ.log("Good: 12 or 34.5 or -6 or -.789");
	IJ.log("Bad: 1.2E1 or 3.45E+1 or -6e0 or -7.89e-1");
	IJ.log("");
	IJ.log("Default values: optimizing=yes stamping=no");
	IJ.log("");
	IJ.log("~~~");
	IJ.log("");
} /* end dumpSyntax */

/*------------------------------------------------------------------*/
private String[] getTokens (
	String options
) {
	final String fileSeparator = System.getProperty("file.separator");
	if (fileSeparator.equals("\\")) {
		options = options.replaceAll("\\\\", "/");
	}
	else {
		options = options.replaceAll(fileSeparator, "/");
	}
	String[] token = new String[0];
	final StringReader sr = new StringReader(options);
	final StreamTokenizer st = new StreamTokenizer(sr);
	st.resetSyntax();
	st.whitespaceChars(0, ' ');
	st.wordChars('!', 255);
	st.quoteChar('\"');
	st.ordinaryChar('=');
	st.parseNumbers();
	final Vector<String> v = new Vector<String>();
	try {
		while (TT_EOF != st.nextToken()) {
			switch (st.ttype) {
				case TT_EOL : {
					break;
				}
				case TT_NUMBER : {
					v.add(new String("" + st.nval));
					break;
				}
				case TT_WORD : {
					v.add(new String(st.sval));
					break;
				}
			}
		}
	} catch (IOException e) {
		IJ.log("IOException exception " + e.getMessage());
		return(token);
	}
	token = v.toArray(token);
	return(token);
} /* end getTokens */

} /* end class Ovuscule_ */

/*====================================================================
|	Ovuscule
\===================================================================*/
class Ovuscule
	implements
		OvusculeSnake2D

{ /* begin class Ovuscule */

/*....................................................................
	private variables
....................................................................*/
private ImageProcessor ip = null;
private double area, RamanujanPerimeter;
private double a11, a12, a22, a33, a3;
private double c1, c2;
private double p1, p2, pq1, pq2, pq;
private double q1, q2, qr1, qr2, qr;
private double r1, r2, rp1, rp2, rp;
private double s1, s2;
private double u1, u2;
private double v1, v2;
private double y1, y2;
private final OvusculeSnake2DNode[] node = new OvusculeSnake2DNode[3];
private int height;
private int width;
private static final double AREA_FACTOR = 2.0 * PI / sqrt(27.0);
private static final double HALF_SQRT2 = sqrt(2.0) / 2.0;
private static final double REGULARIZATION_WEIGHT = 100.0;
private static final double SQRT_TINY =
	sqrt((double)Float.intBitsToFloat((int)0x33FFFFFF));
private static final double SQRT2 = sqrt(2.0);
private static final double SQRT3 = sqrt(3.0);
private static final double THIRD_SQRT2 = sqrt(2.0) / 3.0;

/*....................................................................
	OvusculeSnake2D methods
....................................................................*/
/*------------------------------------------------------------------*/
public double energy (
) {
	final double weightedArea = contrast();
	final double regularization = regularization();
	return(weightedArea + regularization);
} /* end energy */

/*------------------------------------------------------------------*/
public Point2D.Double[] getEnergyGradient (
) {
	final Point2D.Double[] gc = contrastGradient(node[0], node[1]);
	final Point2D.Double[] gr = regularizationGradient(node[0], node[1]);
	return(plus(gc, gr));
} /* end getEnergyGradient */

/*------------------------------------------------------------------*/
public OvusculeSnake2DNode[] getNodes (
) {
	return(node);
} /* end getNodes */

/*------------------------------------------------------------------*/
public OvusculeSnake2DScale[] getScales (
) {
	final int K = (int)ceil(1.0 + RamanujanPerimeter / 2.0);
	if (K > (2 * (width + height))) {
		return(null);
	}
	final int[] s0x1 = new int[K];
	final int[] s0x2 = new int[K];
	final int[] s1x1 = new int[K];
	final int[] s1x2 = new int[K];
	for (int k = 0; (k < K); k++) {
		final double theta = 2.0 * PI * (double)k / (double)K;
		s0x1[k] = (int)round(y1 + c1 * cos(theta) + s1 * sin(theta));
		s0x2[k] = (int)round(y2 + c2 * cos(theta) + s2 * sin(theta));
		s1x1[k] = (int)round(y1
			+ HALF_SQRT2 * (c1 * cos(theta) + s1 * sin(theta)));
		s1x2[k] = (int)round(y2
			+ HALF_SQRT2 * (c2 * cos(theta) + s2 * sin(theta)));
	}
	final OvusculeSnake2DScale[] skin = new OvusculeSnake2DScale[2];
	skin[0] = new OvusculeSnake2DScale(null, null, true, false);
	skin[1] = new OvusculeSnake2DScale(null, null, true, false);
	skin[0].npoints = K;
	skin[0].xpoints = s0x1;
	skin[0].ypoints = s0x2;
	skin[1].npoints = K;
	skin[1].xpoints = s1x1;
	skin[1].ypoints = s1x2;
	return(skin);
} /* end getScales */

/*------------------------------------------------------------------*/
public void setNodes (
	final OvusculeSnake2DNode[] node
) {
	p1 = node[0].x;
	p2 = node[0].y;
	q1 = node[1].x;
	q2 = node[1].y;
	r1 = node[2].x;
	r2 = node[2].y;
	this.node[0].x = p1;
	this.node[0].y = p2;
	this.node[1].x = q1;
	this.node[1].y = q2;
	this.node[2].x = r1;
	this.node[2].y = r2;
	pq1 = p1 - q1;
	pq2 = p2 - q2;
	qr1 = q1 - r1;
	qr2 = q2 - r2;
	rp1 = r1 - p1;
	rp2 = r2 - p2;
	pq = p1 * q2 - p2 * q1;
	qr = q1 * r2 - q2 * r1;
	rp = r1 * p2 - r2 * p1;
	y1 = (p1 + q1 + r1) / 3.0;
	y2 = (p2 + q2 + r2) / 3.0;
	a11 = 3.0 * (p2 * pq2 + q2 * qr2 + r2 * rp2);
	a12 = 3.0 * (p1 * (q2 - 2.0 * p2) + p2 * q1
		+ q1 * (r2 - 2.0 * q2) + q2 * r1 + r1 * (p2 - 2.0 * r2) + r2 * p1);
	a22 = 3.0 * (p1 * pq1 + q1 * qr1 + r1 * rp1);
	a33 = pq + qr + rp;
	a3 = abs(a33);
	c1 = (pq1 - rp1) / 3.0;
	c2 = (pq2 - rp2) / 3.0;
	s1 = qr1 / SQRT3;
	s2 = qr2 / SQRT3;
	final double halfWidth = sqrt(abs(c1 * c1 + s1 * s1));
	final double halfHeight = sqrt(abs(c2 * c2 + s2 * s2));
	u1 = y1 - halfWidth;
	u2 = y2 - halfHeight;
	v1 = y1 + halfWidth;
	v2 = y2 + halfHeight;
	final double a = (a11 + a22) / 3.0;
	final double b = sqrt(abs(a * a - 3.0 * a33 * a33));
	final double semiMinor = THIRD_SQRT2 * sqrt(abs(a - b));
	final double semiMajor = THIRD_SQRT2 * sqrt(abs(a + b));
	area = AREA_FACTOR * a3;
	double excentricity = (semiMajor - semiMinor) / (semiMajor + semiMinor);
	excentricity *= excentricity;
	RamanujanPerimeter = PI * (semiMajor + semiMinor) * (1.0
		+ 3.0 * excentricity / (10.0 + sqrt(abs(4.0 - 3.0 * excentricity))));
} /* end setNodes */

/*....................................................................
	constructor methods
....................................................................*/
/*------------------------------------------------------------------*/
Ovuscule (
	final ImageProcessor ip,
	final double p1,
	final double p2,
	final double q1,
	final double q2,
	final double r1,
	final double r2
) {
	this.ip = ip;
	height = ip.getHeight();
	width = ip.getWidth();
	node[0] = new OvusculeSnake2DNode(p1, p2);
	node[1] = new OvusculeSnake2DNode(q1, q2);
	node[2] = new OvusculeSnake2DNode(r1, r2);
	setNodes(node);
} /* end Ovuscule */

/*....................................................................
	private methods
....................................................................*/
/*------------------------------------------------------------------*/
private double contrast (
) {
	if (area < 1.0) {
		return(1.0 / SQRT_TINY);
	}
	double c = 0.0;
	final int xmin = max((int)floor(u1), 0);
	final int xmax = min((int)ceil(v1), width - 1);
	final int ymin = max((int)floor(u2), 0);
	final int ymax = min((int)ceil(v2), height - 1);
	if ((u1 < xmin) || (xmax < v1) || (u2 < ymin) || (ymax < v2)){
		return(1.0 / SQRT_TINY);
	}
	if ((xmax <= xmin) || (ymax <= ymin)) {
		return(1.0 / SQRT_TINY);
	}
	for (int y = ymin; (y <= ymax); y++) {
		final double dy = y2 - (double)y;
		final double dy2 = dy * dy;
		for (int x = xmin; (x <= xmax); x++) {
			final double dx = y1 - (double)x;
			final double dx2 = dx * dx;
			final double d = sqrt(dx2 + dy2);
			double z = a11 * dx2 + a12 * dx * dy + a22 * dy2;
			if (z < SQRT_TINY) {
				c -= ip.getPixelValue(x, y);
				continue;
			}
			z = a3 / sqrt(z);
			double d0 = (1.0 - z / SQRT2) * d;
			if (d0 < -HALF_SQRT2) {
				c -= ip.getPixelValue(x, y);
				continue;
			}
			if (d0 < HALF_SQRT2) {
				c += SQRT2 * d0 * ip.getPixelValue(x, y);
				continue;
			}
			d0 = (1.0 - z) * d;
			if (d0 < -1.0) {
				c += ip.getPixelValue(x, y);
				continue;
			}
			if (d0 < 1.0) {
				c += (1.0 - d0) * ip.getPixelValue(x, y) / 2.0;
				continue;
			}
		}
	}
	return(c / area);
} /* end contrast */

/*------------------------------------------------------------------*/
private Point2D.Double[] contrastGradient (
	final OvusculeSnake2DNode p,
	final OvusculeSnake2DNode q
) {
	if (area < 1.0) {
		return(null);
	}
	final int xmin = max((int)floor(u1), 0);
	final int xmax = min((int)ceil(v1), width - 1);
	final int ymin = max((int)floor(u2), 0);
	final int ymax = min((int)ceil(v2), height - 1);
	if ((u1 < xmin) || (xmax < v1) || (u2 < ymin) || (ymax < v2)) {
		return(null);
	}
	if ((xmax <= xmin) || (ymax <= ymin)) {
		return(null);
	}
	final Point2D.Double[] gradient = new Point2D.Double[3];
	gradient[0] = new Point2D.Double(0.0, 0.0);
	gradient[1] = new Point2D.Double(0.0, 0.0);
	gradient[2] = new Point2D.Double(0.0, 0.0);
	final double[] z12 =
		{pq1 - qr1, pq2 - qr2, qr1 - rp1, qr2 - rp2, rp1 - pq1, rp2 - pq2};
	for (int y = ymin; (y <= ymax); y++) {
		final double dy = y2 - (double)y;
		final double dy2 = dy * dy;
		for (int x = xmin; (x <= xmax); x++) {
			final double dx = y1 - (double)x;
			final double dx2 = dx * dx;
			final double d = sqrt(dx2 + dy2);
			if (d < SQRT_TINY) {
				continue;
			}
			final double f = ip.getPixelValue(x, y);
			double z = a11 * dx2 + a12 * dx * dy + a22 * dy2;
			if (z < SQRT_TINY) {
				gradient[0].x += f * qr2;
				gradient[0].y -= f * qr1;
				gradient[1].x += f * rp2;
				gradient[1].y -= f * rp1;
				gradient[2].x += f * pq2;
				gradient[2].y -= f * pq1;
				continue;
			}
			z = a3 / sqrt(z);
			double d0 = (1.0 - z / SQRT2) * d;
			if (d0 < -HALF_SQRT2) {
				gradient[0].x += f * qr2;
				gradient[0].y -= f * qr1;
				gradient[1].x += f * rp2;
				gradient[1].y -= f * rp1;
				gradient[2].x += f * pq2;
				gradient[2].y -= f * pq1;
				continue;
			}
			if (d0 < HALF_SQRT2) {
				final double g = SQRT2 * d;
				final double g0 = z * z * z * d / (6.0 * a33);
				final double g1 = a33 * (SQRT2 - z) / (3.0 * g0 * d);
				final double gx = (2.0 * a11 + g1) * dx + a12 * dy;
				final double gy = (2.0 * a22 + g1) * dy + a12 * dx;
				final double h0 = 9.0 * (z12[5] * dx - z12[4] * dy);
				final double h1 = 9.0 * (z12[1] * dx - z12[0] * dy);
				final double h2 = 9.0 * (z12[3] * dx - z12[2] * dy);
				gradient[0].x -= f * (g * qr2 - g0 * (dy * h0 + gx));
				gradient[0].y += f * (g * qr1 - g0 * (dx * h0 - gy));
				gradient[1].x -= f * (g * rp2 - g0 * (dy * h1 + gx));
				gradient[1].y += f * (g * rp1 - g0 * (dx * h1 - gy));
				gradient[2].x -= f * (g * pq2 - g0 * (dy * h2 + gx));
				gradient[2].y += f * (g * pq1 - g0 * (dx * h2 - gy));
				continue;
			}
			d0 = (1.0 - z) * d;
			if (d0 < -1.0) {
				gradient[0].x -= f * qr2;
				gradient[0].y += f * qr1;
				gradient[1].x -= f * rp2;
				gradient[1].y += f * rp1;
				gradient[2].x -= f * pq2;
				gradient[2].y += f * pq1;
				continue;
			}
			if (d0 < 1.0) {
				final double g = (d - 1.0) / 2.0;
				final double g0 = z * z * z * d / (12.0 * a33);
				final double g1 = a33 * (1.0 - z) / (6.0 * g0 * d);
				final double gx = (2.0 * a11 + g1) * dx + a12 * dy;
				final double gy = (2.0 * a22 + g1) * dy + a12 * dx;
				final double h0 = 9.0 * (z12[5] * dx - z12[4] * dy);
				final double h1 = 9.0 * (z12[1] * dx - z12[0] * dy);
				final double h2 = 9.0 * (z12[3] * dx - z12[2] * dy);
				gradient[0].x += f * (g * qr2 - g0 * (dy * h0 + gx));
				gradient[0].y -= f * (g * qr1 - g0 * (dx * h0 - gy));
				gradient[1].x += f * (g * rp2 - g0 * (dy * h1 + gx));
				gradient[1].y -= f * (g * rp1 - g0 * (dx * h1 - gy));
				gradient[2].x += f * (g * pq2 - g0 * (dy * h2 + gx));
				gradient[2].y -= f * (g * pq1 - g0 * (dx * h2 - gy));
				continue;
			}
		}
	}
	final double A = a33 * area;
	gradient[0].x /= A;
	gradient[0].y /= A;
	gradient[1].x /= A;
	gradient[1].y /= A;
	gradient[2].x /= A;
	gradient[2].y /= A;
	return(gradient);
} /* end contrastGradient */

/*------------------------------------------------------------------*/
private Point2D.Double[] plus (
	final Point2D.Double[] gc,
	final Point2D.Double[] gr
) {
	if ((null == gc) || (null == gr)) {
		return(null);
	}
	final int K = gc.length;
	if (K != gr.length) {
		return(null);
	}
	final Point2D.Double[] g = new Point2D.Double[K];
	for (int k = 0; (k < K); k++) {
		g[k] = new Point2D.Double(gc[k].x + gr[k].x, gc[k].y + gr[k].y);
	}
	return(g);
} /* end plus */

/*------------------------------------------------------------------*/
private double regularization (
) {
	final double regularization = min(min(pq2 * pq2, qr2 * qr2), rp2 * rp2);
	return(REGULARIZATION_WEIGHT * regularization / area);
} /* end regularization */

/*------------------------------------------------------------------*/
private Point2D.Double[] regularizationGradient (
	final OvusculeSnake2DNode p,
	final OvusculeSnake2DNode q
) {
	final Point2D.Double[] gradient = new Point2D.Double[3];
	gradient[0] = new Point2D.Double(0.0, 0.0);
	gradient[1] = new Point2D.Double(0.0, 0.0);
	gradient[2] = new Point2D.Double(0.0, 0.0);
	final double ppqq2 = pq2 * pq2;
	final double qqrr2 = qr2 * qr2;
	final double rrpp2 = rp2 * rp2;
	double A = 0.0;
	if ((ppqq2 <= qqrr2) && (ppqq2 <= rrpp2)) {
		gradient[0].x = -pq2 * qr2;
		gradient[0].y = 2.0 * a33 + pq2 * qr1;
		gradient[1].x = -rp2 * pq2;
		gradient[1].y = -2.0 * a33 + rp1 * pq2;
		gradient[2].x = -ppqq2;
		gradient[2].y = pq1 * pq2;
		A = REGULARIZATION_WEIGHT * pq2 / (a33 * area);
	}
	else if ((qqrr2 <= rrpp2) && (qqrr2 <= ppqq2)) {
		gradient[0].x = -qqrr2;
		gradient[0].y = qr1 * qr2;
		gradient[1].x = -qr2 * rp2;
		gradient[1].y = 2.0 * a33 + qr2 * rp1;
		gradient[2].x = -pq2 * qr2;
		gradient[2].y = -2.0 * a33 + pq1 * qr2;
		A = REGULARIZATION_WEIGHT * qr2 / (a33 * area);
	}
	else if ((rrpp2 <= ppqq2) && (rrpp2 <= qqrr2)) {
		gradient[0].x = -qr2 * rp2;
		gradient[0].y = -2.0 * a33 + qr1 * rp2;
		gradient[1].x = -rrpp2;
		gradient[1].y = rp1 * rp2;
		gradient[2].x = -rp2 * pq2;
		gradient[2].y = 2.0 * a33 + rp2 * pq1;
		A = REGULARIZATION_WEIGHT * rp2 / (a33 * area);
	}
	gradient[0].x *= A;
	gradient[0].y *= A;
	gradient[1].x *= A;
	gradient[1].y *= A;
	gradient[2].x *= A;
	gradient[2].y *= A;
	return(gradient);
} /* end regularizationGradient */

} /* end class Ovuscule */
